@@ -20,7 +20,7 @@ class Redis
     def initialize(node_configs, options = {})
       @tag = options[:tag] || /^\{(.+?)\}/
       @ring = options[:ring] || HashRing.new
-      @node_configs = node_configs.dup
+      @node_configs = node_configs.map(&:dup)
       @default_options = options.dup
       node_configs.each { |node_config| add_node(node_config) }
       @subscribed_node = nil
@@ -41,6 +41,8 @@ class Redis
     def add_node(options)
       options = { url: options } if options.is_a?(String)
       options = @default_options.merge(options)
+      options.delete(:tag)
+      options.delete(:ring)
       @ring.add_node Redis.new(options)
     end
 
@@ -64,6 +66,10 @@ class Redis
       on_each_node :quit
     end
 
+    def close
+      on_each_node :close
+    end
+
     # Asynchronously save the dataset to disk.
     def bgsave
       on_each_node :bgsave
@@ -124,6 +130,11 @@ class Redis
       node_for(key).expireat(key, unix_time, **kwargs)
     end
 
+    # Get the expiration for a key as a UNIX timestamp.
+    def expiretime(key)
+      node_for(key).expiretime(key)
+    end
+
     # Get the time to live (in seconds) for a key.
     def ttl(key)
       node_for(key).ttl(key)
@@ -139,6 +150,11 @@ class Redis
       node_for(key).pexpireat(key, ms_unix_time, **kwarg)
     end
 
+    # Get the expiration for a key as number of milliseconds from UNIX Epoch.
+    def pexpiretime(key)
+      node_for(key).pexpiretime(key)
+    end
+
     # Get the time to live (in milliseconds) for a key.
     def pttl(key)
       node_for(key).pttl(key)
@@ -161,6 +177,7 @@ class Redis
 
     # Delete a key.
     def del(*args)
+      args.flatten!(1)
       keys_per_node = args.group_by { |key| node_for(key) }
       keys_per_node.inject(0) do |sum, (node, keys)|
         sum + node.del(*keys)
@@ -169,6 +186,7 @@ class Redis
 
     # Unlink keys.
     def unlink(*args)
+      args.flatten!(1)
       keys_per_node = args.group_by { |key| node_for(key) }
       keys_per_node.inject(0) do |sum, (node, keys)|
         sum + node.unlink(*keys)
@@ -177,23 +195,16 @@ class Redis
 
     # Determine if a key exists.
     def exists(*args)
-      if !Redis.exists_returns_integer && args.size == 1
-        ::Redis.deprecate!(
-          "`Redis#exists(key)` will return an Integer in redis-rb 4.3, if you want to keep the old behavior, " \
-          "use `exists?` instead. To opt-in to the new behavior now you can set Redis.exists_returns_integer = true. " \
-          "(#{::Kernel.caller(1, 1).first})\n"
-        )
-        exists?(*args)
-      else
-        keys_per_node = args.group_by { |key| node_for(key) }
-        keys_per_node.inject(0) do |sum, (node, keys)|
-          sum + node._exists(*keys)
-        end
+      args.flatten!(1)
+      keys_per_node = args.group_by { |key| node_for(key) }
+      keys_per_node.inject(0) do |sum, (node, keys)|
+        sum + node.exists(*keys)
       end
     end
 
     # Determine if any of the keys exists.
     def exists?(*args)
+      args.flatten!(1)
       keys_per_node = args.group_by { |key| node_for(key) }
       keys_per_node.each do |node, keys|
         return true if node.exists?(*keys)
@@ -297,7 +308,7 @@ class Redis
     end
 
     # Set multiple keys to multiple values.
-    def mset(*_args)
+    def mset(*)
       raise CannotDistribute, :mset
     end
 
@@ -306,7 +317,7 @@ class Redis
     end
 
     # Set multiple keys to multiple values, only if none of the keys exist.
-    def msetnx(*_args)
+    def msetnx(*)
       raise CannotDistribute, :msetnx
     end
 
@@ -331,11 +342,13 @@ class Redis
 
     # Get the values of all the given keys as an Array.
     def mget(*keys)
+      keys.flatten!(1)
       mapped_mget(*keys).values_at(*keys)
     end
 
     # Get the values of all the given keys as a Hash.
     def mapped_mget(*keys)
+      keys.flatten!(1)
       keys.group_by { |k| node_for k }.inject({}) do |results, (node, subkeys)|
         results.merge! node.mapped_mget(*subkeys)
       end
@@ -367,20 +380,21 @@ class Redis
     end
 
     # Count the number of set bits in a range of the string value stored at key.
-    def bitcount(key, start = 0, stop = -1)
-      node_for(key).bitcount(key, start, stop)
+    def bitcount(key, start = 0, stop = -1, scale: nil)
+      node_for(key).bitcount(key, start, stop, scale: scale)
     end
 
     # Perform a bitwise operation between strings and store the resulting string in a key.
     def bitop(operation, destkey, *keys)
+      keys.flatten!(1)
       ensure_same_node(:bitop, [destkey] + keys) do |node|
-        node.bitop(operation, destkey, *keys)
+        node.bitop(operation, destkey, keys)
       end
     end
 
     # Return the position of the first bit set to 1 or 0 in a string.
-    def bitpos(key, bit, start = nil, stop = nil)
-      node_for(key).bitpos(key, bit, start, stop)
+    def bitpos(key, bit, start = nil, stop = nil, scale: nil)
+      node_for(key).bitpos(key, bit, start, stop, scale: scale)
     end
 
     # Set the string value of a key and return its old value.
@@ -463,23 +477,15 @@ class Redis
       timeout = if args.last.is_a?(Hash)
         options = args.pop
         options[:timeout]
-      elsif args.last.respond_to?(:to_int)
-        last_arg = args.pop
-        ::Redis.deprecate!(
-          "Passing the timeout as a positional argument is deprecated, it should be passed as a keyword argument:\n" \
-          "  redis.#{cmd}(#{args.map(&:inspect).join(', ')}, timeout: #{last_arg.to_int})" \
-          "(called from: #{caller(2, 1).first})"
-        )
-        last_arg.to_int
       end
 
-      keys = args.flatten
+      args.flatten!(1)
 
-      ensure_same_node(cmd, keys) do |node|
+      ensure_same_node(cmd, args) do |node|
         if timeout
-          node.__send__(cmd, keys, timeout: timeout)
+          node.__send__(cmd, args, timeout: timeout)
         else
-          node.__send__(cmd, keys)
+          node.__send__(cmd, args)
         end
       end
     end
@@ -490,6 +496,18 @@ class Redis
       _bpop(:blpop, args)
     end
 
+    def bzpopmax(*args)
+      _bpop(:bzpopmax, args) do |reply|
+        reply.is_a?(Array) ? [reply[0], reply[1], Floatify.call(reply[2])] : reply
+      end
+    end
+
+    def bzpopmin(*args)
+      _bpop(:bzpopmin, args) do |reply|
+        reply.is_a?(Array) ? [reply[0], reply[1], Floatify.call(reply[2])] : reply
+      end
+    end
+
     # Remove and get the last element in a list, or block until one is
     # available.
     def brpop(*args)
@@ -498,9 +516,9 @@ class Redis
 
     # Pop a value from a list, push it to another list and return it; or block
     # until one is available.
-    def brpoplpush(source, destination, deprecated_timeout = 0, **options)
+    def brpoplpush(source, destination, **options)
       ensure_same_node(:brpoplpush, [source, destination]) do |node|
-        node.brpoplpush(source, destination, deprecated_timeout, **options)
+        node.brpoplpush(source, destination, **options)
       end
     end
 
@@ -534,29 +552,43 @@ class Redis
       node_for(key).ltrim(key, start, stop)
     end
 
+    # Iterate over keys, blocking and removing elements from the first non empty liist found.
+    def blmpop(timeout, *keys, modifier: "LEFT", count: nil)
+      ensure_same_node(:blmpop, keys) do |node|
+        node.blmpop(timeout, *keys, modifier: modifier, count: count)
+      end
+    end
+
+    # Iterate over keys, removing elements from the first non list found.
+    def lmpop(*keys, modifier: "LEFT", count: nil)
+      ensure_same_node(:lmpop, keys) do |node|
+        node.lmpop(*keys, modifier: modifier, count: count)
+      end
+    end
+
     # Get the number of members in a set.
     def scard(key)
       node_for(key).scard(key)
     end
 
     # Add one or more members to a set.
-    def sadd(key, member)
-      node_for(key).sadd(key, member)
+    def sadd(key, *members)
+      node_for(key).sadd(key, *members)
     end
 
     # Add one or more members to a set.
-    def sadd?(key, member)
-      node_for(key).sadd?(key, member)
+    def sadd?(key, *members)
+      node_for(key).sadd?(key, *members)
     end
 
     # Remove one or more members from a set.
-    def srem(key, member)
-      node_for(key).srem(key, member)
+    def srem(key, *members)
+      node_for(key).srem(key, *members)
     end
 
     # Remove one or more members from a set.
-    def srem?(key, member)
-      node_for(key).srem?(key, member)
+    def srem?(key, *members)
+      node_for(key).srem?(key, *members)
     end
 
     # Remove and return a random member from a set.
@@ -603,43 +635,49 @@ class Redis
 
     # Subtract multiple sets.
     def sdiff(*keys)
+      keys.flatten!(1)
       ensure_same_node(:sdiff, keys) do |node|
-        node.sdiff(*keys)
+        node.sdiff(keys)
       end
     end
 
     # Subtract multiple sets and store the resulting set in a key.
     def sdiffstore(destination, *keys)
-      ensure_same_node(:sdiffstore, [destination] + keys) do |node|
-        node.sdiffstore(destination, *keys)
+      keys.flatten!(1)
+      ensure_same_node(:sdiffstore, [destination].concat(keys)) do |node|
+        node.sdiffstore(destination, keys)
       end
     end
 
     # Intersect multiple sets.
     def sinter(*keys)
+      keys.flatten!(1)
       ensure_same_node(:sinter, keys) do |node|
-        node.sinter(*keys)
+        node.sinter(keys)
       end
     end
 
     # Intersect multiple sets and store the resulting set in a key.
     def sinterstore(destination, *keys)
-      ensure_same_node(:sinterstore, [destination] + keys) do |node|
-        node.sinterstore(destination, *keys)
+      keys.flatten!(1)
+      ensure_same_node(:sinterstore, [destination].concat(keys)) do |node|
+        node.sinterstore(destination, keys)
       end
     end
 
     # Add multiple sets.
     def sunion(*keys)
+      keys.flatten!(1)
       ensure_same_node(:sunion, keys) do |node|
-        node.sunion(*keys)
+        node.sunion(keys)
       end
     end
 
     # Add multiple sets and store the resulting set in a key.
     def sunionstore(destination, *keys)
-      ensure_same_node(:sunionstore, [destination] + keys) do |node|
-        node.sunionstore(destination, *keys)
+      keys.flatten!(1)
+      ensure_same_node(:sunionstore, [destination].concat(keys)) do |node|
+        node.sunionstore(destination, keys)
       end
     end
 
@@ -680,6 +718,20 @@ class Redis
       node_for(key).zmscore(key, *members)
     end
 
+    # Iterate over keys, blocking and removing members from the first non empty sorted set found.
+    def bzmpop(timeout, *keys, modifier: "MIN", count: nil)
+      ensure_same_node(:bzmpop, keys) do |node|
+        node.bzmpop(timeout, *keys, modifier: modifier, count: count)
+      end
+    end
+
+    # Iterate over keys, removing members from the first non empty sorted set found.
+    def zmpop(*keys, modifier: "MIN", count: nil)
+      ensure_same_node(:zmpop, keys) do |node|
+        node.zmpop(*keys, modifier: modifier, count: count)
+      end
+    end
+
     # Return a range of members in a sorted set, by index, score or lexicographical ordering.
     def zrange(key, start, stop, **options)
       node_for(key).zrange(key, start, stop, **options)
@@ -738,43 +790,49 @@ class Redis
 
     # Get the intersection of multiple sorted sets
     def zinter(*keys, **options)
+      keys.flatten!(1)
       ensure_same_node(:zinter, keys) do |node|
-        node.zinter(*keys, **options)
+        node.zinter(keys, **options)
       end
     end
 
     # Intersect multiple sorted sets and store the resulting sorted set in a new
     # key.
-    def zinterstore(destination, keys, **options)
-      ensure_same_node(:zinterstore, [destination] + keys) do |node|
+    def zinterstore(destination, *keys, **options)
+      keys.flatten!(1)
+      ensure_same_node(:zinterstore, [destination].concat(keys)) do |node|
         node.zinterstore(destination, keys, **options)
       end
     end
 
     # Return the union of multiple sorted sets.
     def zunion(*keys, **options)
+      keys.flatten!(1)
       ensure_same_node(:zunion, keys) do |node|
-        node.zunion(*keys, **options)
+        node.zunion(keys, **options)
       end
     end
 
     # Add multiple sorted sets and store the resulting sorted set in a new key.
-    def zunionstore(destination, keys, **options)
-      ensure_same_node(:zunionstore, [destination] + keys) do |node|
+    def zunionstore(destination, *keys, **options)
+      keys.flatten!(1)
+      ensure_same_node(:zunionstore, [destination].concat(keys)) do |node|
         node.zunionstore(destination, keys, **options)
       end
     end
 
     # Return the difference between the first and all successive input sorted sets.
     def zdiff(*keys, **options)
+      keys.flatten!(1)
       ensure_same_node(:zdiff, keys) do |node|
-        node.zdiff(*keys, **options)
+        node.zdiff(keys, **options)
       end
     end
 
     # Compute the difference between the first and all successive input sorted sets
     # and store the resulting sorted set in a new key.
-    def zdiffstore(destination, keys, **options)
+    def zdiffstore(destination, *keys, **options)
+      keys.flatten!(1)
       ensure_same_node(:zdiffstore, [destination] + keys) do |node|
         node.zdiffstore(destination, keys, **options)
       end
@@ -801,7 +859,7 @@ class Redis
     end
 
     def mapped_hmset(key, hash)
-      node_for(key).hmset(key, *hash.to_a.flatten)
+      node_for(key).hmset(key, hash)
     end
 
     # Get the value of a hash field.
@@ -811,11 +869,13 @@ class Redis
 
     # Get the values of all the given hash fields.
     def hmget(key, *fields)
-      node_for(key).hmget(key, *fields)
+      fields.flatten!(1)
+      node_for(key).hmget(key, fields)
     end
 
     def mapped_hmget(key, *fields)
-      Hash[*fields.zip(hmget(key, *fields)).flatten]
+      fields.flatten!(1)
+      node_for(key).mapped_hmget(key, fields)
     end
 
     def hrandfield(key, count = nil, **options)
@@ -824,7 +884,8 @@ class Redis
 
     # Delete one or more hash fields.
     def hdel(key, *fields)
-      node_for(key).hdel(key, *fields)
+      fields.flatten!(1)
+      node_for(key).hdel(key, fields)
     end
 
     # Determine if a hash field exists.
@@ -881,7 +942,7 @@ class Redis
 
     # Stop listening for messages posted to the given channels.
     def unsubscribe(*channels)
-      raise "Can't unsubscribe if not subscribed." unless subscribed?
+      raise SubscriptionError, "Can't unsubscribe if not subscribed." unless subscribed?
 
       @subscribed_node.unsubscribe(*channels)
     end
@@ -928,9 +989,7 @@ class Redis
     def multi(&block)
       raise CannotDistribute, :multi unless @watch_key
 
-      result = node_for(@watch_key).multi(&block)
-      @watch_key = nil if block_given?
-      result
+      node_for(@watch_key).multi(&block)
     end
 
     # Execute all commands issued after MULTI.
@@ -1020,7 +1079,8 @@ class Redis
     end
 
     def key_tag(key)
-      key.to_s[@tag, 1] if @tag
+      key = key.to_s
+      key[@tag, 1] if key.match?(@tag)
     end
 
     def ensure_same_node(command, keys)
